{-# LANGUAGE TupleSections #-}

module Extract where

{- extract to Fomega

Examples:
---------

MiniAgda

  data Vec (A : Set) : Nat -> Set
  { vnil  : Vec A zero
  ; vcons : [n : Nat] -> (head : A) -> (tail : Vec A n) -> Vec A (suc n) 
  } fields head, tail

  fun length : [A : Set] -> [n : Nat] -> Vec A n -> <n : Nat>
  { length .A .zero    (vnil A)         = zero
  ; length .A .(suc n) (vcons A n a as) = suc (length A n as)
  } 

Fomega

  data Vec (A : Set) : Set
  { vnil  : Vec A
  ; vcons : (head : A) -> (tail : Vec A) -> Vec A
  } 

  fun head : [A : Set] -> Vec A -> A
  { head (vcons 'head 'tail) = 'head
  }

  fun tail : [A : Set] -> Vec A -> A
  { head (vcons 'head 'tail) = 'tail
  }

  fun length : [A : Set] -> Vec A -> Nat
  { length [A]  vnil             = zero
  ; length [A] (vcons [.A] a as) = suc (length [A] as)
  } 


Bidirectional extraction
========================

Types 

  Base ::= D As         data type
         | ?            inexpressible type 
                        
  A,B ::= Base | A -> B | [x:K] -> B | [] -> B  with erasure markers
  A0, B0 ::= Base | A0 -> B0 | [x:K0] -> B0     without erasure markers

  |.| erase erasure markers

Inference mode:

  Term extraction:  Gamma |- t :> A  --> e    |Gamma| |- e : |A|
  Type extraction:  Gamma |- T :> K  --> A    |Gamma| |- A : |K|
  Kind extraction:  Gamma |- U :> [] --> K    |Gamma| |- K : []

Checking mode:

  Term extraction:  Gamma |- t <: A  --> e    |Gamma| |- e : |A|
  Type extraction:  Gamma |- T <: K  --> A    |Gamma| |- A : |K|
  Kind extraction:  Gamma |- U <: [] --> K    |Gamma| |- K : []

Type and kind extraction keep erasure markers!
  
Checking abstraction:

  Relevant abstraction:
  Gamma, x:A |- t <: B --> e
  --------------------------------
  Gamma |- \x.t <: A -> B --> \x.e

  Type abstraction:
  Gamma, x:K |- t <: B --> e : B0
  ----------------------------------------
  Gamma |- \[x].t <: [x:K] -> B --> \[x].e
      also \xt

  Irrelevant abstraction:
  Gamma |- t : B --> e
  -------------------------------
  Gamma |- \[x].t : [] -> B --> e
      also \xt

  Relevant abstraction at unknown type:
  Gamma, x:? |- t : ? --> e
  --------------------------
  Gamma |- \x.t : ? --> \x.e
  
  Irrelevant abstraction at unknown type:
  Gamma |- t : ? --> e
  -------------------------
  Gamma |- \[x].t : ? --> e
  
Checking by inference:
  
  Gamma |- t :> A --> e    e : |A| <: |B| --> e'
  ----------------------------------------------
  Gamma |- t <: B --> e' : B0 

Casting:

  ------------------ A0 does not contain ?
  e : A0 <: A0 --> e

  ----------------------- A0 != B0 or one does contain ?
  e : A0 <: B0 --> cast e

Inferring variable:
  
  ----------------------------
  Gamma |- x :> Gamma(x) --> x

Inferring application:

  Relevant application:
  Gamma |- t :> A -> B --> f     Gamma |- u <: A --> e
  ----------------------------------------------------
  Gamma |- t u :> B --> f e

  Type application:
  Gamma |- t :> [x:K] -> B --> f   Gamma |- u <: K --> A
  ------------------------------------------------------
  Gamma |- t [u] :> : B[A/x] --> f [A] 
      also  t u 
  
  Irrelevant application:
  Gamma |- t :> [] -> B --> f
  ---------------------------
  Gamma |- t [u] :> B --> f
      also  t u

  Relevant application at unknown type:
  Gamma |- t :> ? --> f     Gamma |- u <: ? --> e
  -----------------------------------------------
  Gamma |- t u :> ? --> f e

  Irrelevant application at unknown type:
  Gamma |- t :> ? --> f
  -------------------------
  Gamma |- t [u] :> ? --> f



-}

import Char

import Control.Applicative
import Control.Monad
import Control.Monad.Error
import Control.Monad.Reader
import Control.Monad.Writer
import Control.Monad.State

import Data.Traversable (Traversable) 
import qualified Data.Traversable as Traversable
import Data.Map (Map)
import qualified Data.Map as Map

import Abstract
import Value
import Eval
import TCM
import TraceError


runExtract sig k = runErrorT (runReaderT (runStateT k (initWithSig sig)) emptyContext)

-- extraction

type FExpr        = Expr
type FDeclaration = Declaration
type FClause      = Clause
type FPattern     = Pattern
type FConstructor = Constructor
type FTypeSig     = TypeSig
type FFun         = Fun
type FTelescope   = Telescope

type FTVal        = TVal

extractDecls :: [EDeclaration] -> TypeCheck [FDeclaration]
extractDecls ds = concat <$> mapM extractDecl ds

extractDecl :: EDeclaration -> TypeCheck [FDeclaration]
extractDecl d = 
  case d of
    MutualDecl _ ds -> return [] -- TODO! extractDecls ds
    OverrideDecl{} -> fail $ "extractDecls internal error: overrides impossible"
    MutualFunDecl _ _ funs -> extractFuns funs
    FunDecl _ fun -> extractFun fun 
    LetDecl evl ts e -> extractLet evl ts e
    DataDecl n _ co _ tel ty cs fields -> extractDataDecl n co tel ty cs

extractFuns :: [Fun] -> TypeCheck [FDeclaration]
extractFuns funs = concat <$> mapM extractFun funs

extractFun :: Fun -> TypeCheck [FDeclaration]
extractFun fun = return []

extractLet :: Bool -> TypeSig -> Expr -> TypeCheck [FDeclaration]
extractLet evl (TypeSig n t) e = do
  t <- extractType =<< whnf' t
  e <- extractCheck e =<< whnf' t
  return [LetDecl evl (TypeSig n t) e] 

extractDataDecl :: Name -> Co -> Telescope -> Type -> [Constructor] -> TypeCheck [FDeclaration]
extractDataDecl n co tel ty cs = do
  -- k    <- extrTyp <$> lookupSymb n
  tel' <- extractKindTel tel
  Just core <- addBinds tel $ extractKind =<< whnf' ty 
  -- (_, core) = typeToTele' (length tel') k 
  cs   <- mapM (extractConstructor tel) cs  
  return [DataDecl n NotSized co [] tel' core cs []]
  
extractConstructor :: Telescope -> Constructor -> TypeCheck FConstructor
extractConstructor tel (TypeSig n t) = do
  -- compute full extracted constructor type and add to the signature
  t' <- extractType =<< whnf emptyEnv (teleToTypeErase tel t) 
  setExtrTyp n t'
  let (tel',core) = typeToTele' (length tel) t'
  return $ TypeSig n core
  -- compute type minus telescope
  -- TypeSig n <$> (extractType =<< whnf' t)


-- extraction a term from a term -------------------------------------

extractInfer :: Expr -> TypeCheck (FExpr, FTVal)
extractInfer e = do
  case e of

    Var x -> (Var x,) . typ . domain <$> lookupName1 x

    App f e0 -> do
      let (er, e) = isErasedExpr e0
      (f, tv) <- extractInfer f
      fv <- funView tv
      case fv of
        EraseArg bv -> return (f,bv)
        Forall x dom env b -> do 
          e <- extractTypeAt e (typ dom)
          bv <- (\ xv -> whnf (update env x xv) b) =<< whnf' e 
          return $ (App f (erasedExpr e), bv) 
        Arrow av bv -> return (if er then f else App f e, bv)

    -- TODO: other expressions!

    _ -> return (Irr, VIrr)

extractCheck :: Expr -> FTVal -> TypeCheck (FExpr)
extractCheck e tv = do
  case e of
    Lam dec y e -> do
      fv <- funView tv
      case fv of
        EraseArg bv        -> extractCheck e bv -- discard lambda
        Forall x dom env b -> 
          Lam (decor dom) y <$> do
            newWithGen y dom $ \ i _ -> 
              extractCheck e =<< whnf (update env x (VGen i)) b
        Arrow av bv        ->
          if erased dec then extractCheck e bv
           else Lam dec y <$> do
             new' y (defaultDomain av) $
               extractCheck e bv

    -- TODO: let, case

    _ -> do
      (e,tv') <- extractInfer e
      insertCast e tv tv'

insertCast :: FExpr -> FTVal -> FTVal -> TypeCheck FExpr
insertCast e tv1 tv2 = loop tv1 tv2 where
  loop tv1 tv2 =
    case (tv1,tv2) of
      (VIrr,_) -> return $ castExpr e
      (_,VIrr) -> return $ castExpr e
      _  -> return e -- TODO!
 
funView :: FTVal -> TypeCheck FunView
funView tv = 
  case tv of
    -- erasure mark
    VPi x dom env e | erased (decor dom) && typ dom == VIrr -> 
      EraseArg <$> whnf (update env x VIrr) e
    -- forall
    VPi x dom env e | erased (decor dom) -> 
      return $ Forall x dom env e
    -- function type
    VPi x dom env e ->
      Arrow (typ dom) <$> whnf (update env x VIrr) e
    -- any other type can be a function type
    _ -> return $ Arrow VIrr VIrr

data FunView 
  = Arrow    FTVal FTVal            -- A -> B
  | Forall   Name Domain Env FType  -- forall X:K. A
  | EraseArg FTVal                  -- [] -> B
  

-- extracting a kind from a value ------------------------------------

type FKind = Expr -- FKind ::= Set | FKind -> FKind | [Irr] -> FKind

star :: FKind
star = Sort $ Set Zero

extractSet :: Sort Val -> Maybe FKind
extractSet s = 
  case s of
    SortC _ -> Nothing
    Set _   -> Just $ star
    CoSet _ -> Just $ star

-- keep irrelevant entries
extractKindTel :: Telescope -> TypeCheck FTelescope
extractKindTel [] = return []
extractKindTel (TBind x dom : tel) = do
  dom  <- Traversable.mapM whnf' dom
  dom' <- extractKindDom dom 
  if erased (decor dom') then
    newIrr x $  
      (TBind x dom' :) <$> extractKindTel tel
   else newTyVar x (typ dom') $ \ i ->
      (TBind (tyVarName i) dom' :) <$> extractKindTel tel 

{-
-- keep irrelevant entries
extractKindTel :: Telescope -> TypeCheck FTelescope
extractKindTel tel = do
  tv     <- whnf' (teleToType tel star)
  Just k <- extractKind tv
  let (tel, s) = typeToTele k
  return tel
  -- throw away erasure marks
  -- return $ filter (\ tb -> not $ erased $ decor $ boundDom tb) tel
-}

extractKindDom :: Domain -> TypeCheck (Dom FKind)
extractKindDom dom = 
  maybe (defaultIrrDom Irr) defaultDomain <$>
    if erased (decor dom) then return Nothing 
     else extractKind (typ dom)

extractKind :: TVal -> TypeCheck (Maybe FKind)
extractKind tv = 
  case tv of
    VSort s -> return $ extractSet s
    VPi x dom env b -> new' x dom $ do
       bv  <- whnf (update env x VIrr) b
       mk' <- extractKind bv
       case mk' of
         Nothing -> return Nothing     
         Just k' -> do
           dom' <- extractKindDom dom
           return $ Just $ Pi (TBind "" dom') k' 
    _ -> return Nothing

-- extracting a type constructor from a value ------------------------

type FType = Expr 
{- FType ::= Irr                 -- not expressible in Fomega
           | D FTypes            -- data type
           | X FTypes            -- type variable
           | FType -> FType      -- function type
           | [X:FKind] -> FType  -- polymorphic type
           | [Irr] -> FType      -- erasure marker
 -}

tyVarName i = "a" ++ show i

newTyVar :: Name -> FKind -> (Int -> TypeCheck a) -> TypeCheck a
newTyVar x k cont = newWithGen x (defaultDomain (VClos emptyEnv k)) $ 
  \ i _ -> cont i                  -- store kinds unevaluated

addFKindTel :: FTelescope -> TypeCheck a -> TypeCheck a
addFKindTel [] cont = cont
addFKindTel (TBind x dom : tel) cont = newTyVar x (typ dom) $ const $ 
  addFKindTel tel cont 

extractType :: TVal -> TypeCheck FType
extractType = extractTypeAt star

extractTypeAt :: FKind -> TVal -> TypeCheck FType
extractTypeAt k tv = do
  case (tv,k) of

    -- relevant function space --> non-dependent
    (VPi x dom env b,_) | not (erased (decor dom)) -> do
      a <- extractType (typ dom) 
      -- new' x dom $ do
      bv <- whnf (update env x VIrr) b
      b  <- extractType bv 
      return $ Pi (TBind "" (defaultDomain a)) b

    -- irrelevant function space --> forall or erasure marker  
    (VPi x dom env b,_) | erased (decor dom) -> do
      mk <- extractKind (typ dom)
      case mk of
        Nothing -> do -- new' x dom $ do
          bv <- whnf (update env x VIrr) b
          b  <- extractType bv 
          return $ Pi (TBind "" (defaultIrrDom Irr)) b
        Just k' -> do
          newTyVar x k' $ \ i -> do
            bv <- whnf (update env x (VGen i)) b
            b  <- extractType bv 
            return $ Pi (TBind (tyVarName i) (defaultIrrDom k')) b 

    (VApp (VDef (DefId Dat n)) vs, _) -> do
      k  <- extrTyp <$> lookupSymb n  -- get kind of dname from signature
      as <- extractTypes k vs  -- turn vs into types as at kind k
      return $ foldl App (Def (DefId Dat n)) as

    (VGen i,_) -> do     
      VClos _ k <- (typ . fromOne . domain) <$> lookupGen i  -- get kind of var from cxt
      return $ Var (tyVarName i)

    (VApp (VGen i) vs,_) -> do     
      VClos _ k <- (typ . fromOne . domain) <$> lookupGen i  -- get kind of var from cxt
      as <- extractTypes k vs  -- turn vs into types as at kind k
      return $ foldl App (Var (tyVarName i)) as

    (VLam x env e, Pi (TBind _ dom) k) | erased (decor dom) -> do 
      tv <- whnf (update env x VIrr) e
      extractTypeAt k tv 
 
    (VLam x env e, Pi (TBind _ dom) k) -> newTyVar x (typ dom) $ \ i -> do
      tv <- whnf (update env x (VGen i)) e
      Lam defaultDec (tyVarName i) <$> extractTypeAt k tv 
 
    (VLam{},_) -> error $ "panic! extractTypeAt " ++ show (tv,k)
 
    (VSing _ tv,_) -> extractTypeAt k tv 
 
    (VUp v _,_)    -> extractTypeAt k v 
 
    _ -> return Irr
 
extractTypes :: FKind -> [TVal] -> TypeCheck [FType]
extractTypes k vs = 
  case (k,vs) of
    (_, []) -> return []
    (Pi (TBind _ dom) k, v:vs) | erased (decor dom) -> extractTypes k vs
    (Pi (TBind _ dom) k, v:vs) -> do
      v  <- whnfClos v
      a  <- extractTypeAt (typ dom) v
      as <- extractTypes k vs
      return $ a : as
    _ -> error $ "panic! extractTypes  " ++ show k ++ "  " ++ show vs

